/* ============================= C MeatAxe ==================================
   File:        $Id: kernel-0.c,v 1.2 2003/04/09 18:12:25 hoffmant Exp $
   Comment:     Finite field arithmetic and common functions. `Small' version
		for field orders q <= 256. Originally based on the `hprout.c'
		written by Klaus Lux.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   Modified by Tom Hoffman, Mathematics Department, University of
   Arizona, USA <hoffmant@math.arizona.edu>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"


/* ------------------------------------------------------------------
   Gobal data
   ------------------------------------------------------------------ */

MTX_DEFINE_FILE_INFO

typedef unsigned char BYTE;
static int MPB = 0;		/* No. of marks per byte */
static int LPR = 0;		/* Long ints per row */





/* ------------------------------------------------------------------
   Argument checking (for debbugging)
   ------------------------------------------------------------------ */

#if defined(_DEBUG)
#define CHECKRANGE(x,lo,hi) if ((x)<(lo)||(x)>(hi)) {\
	MtxError(&Mtx_ThisFile,__LINE__,\
	"ZZZ RANGE CHECK ERROR: %d <= %d <= %d ?\n",\
	(int)(lo),(int)(x),(int)(hi));}
#else
#define CHECKRANGE(x,lo,hi)
#endif

#define CHECKFEL(x) CHECKRANGE(x,0,FfOrder-1)
#define CHECKCOL(x)  CHECKRANGE(x,1,FfNoc)



/*{{{ --------------------------------------------------------------------------
!section kernel.ff.ff
!variable FfCurrentRowSize "Row size"
!synopsis
    extern size_t FfCurrentRowSize;
!description
    This variable contains the size of a single row in memory. Its value
    is always equal to |FfRowSize(FfNoc)|.
!seealso  FfRowSize
-------------------------------------------------------------------------- }}}*/

size_t FfCurrentRowSize = (size_t) -1;



/*{{{ --------------------------------------------------------------------------
!variable FfCurrentRowSizeIo "I/O row size"
!synopsis
    extern int FfCurrentRowSizeIo;
!description
    This variable contains the number of bytes used by a single row. This
    is exactly the number oytes occupied by a row if it is stored in a data
    file. In memory, however, rows are padded with extra bytes to make the
    row size a multiple of sizeof(long).
!seealso  FfCurrentRowSize
-------------------------------------------------------------------------- }}}*/

int FfCurrentRowSizeIo = -1;



/*{{{ --------------------------------------------------------------------------
!function FfAdd "Finite field addition"
!synopsis
    FEL FfAdd(FEL a, FEL b);
!arg a
    First field element.
!arg b
    Second field element.
!return
    $ |a| + |b| $
!description
    This function returns the sum of two field elements. Before you can
    use this function, the field must have been selected with |FfSetField()|.
    The arguments are not checked. If either argument is not in the allowed
    range the result is undefined and the program may crash.

    |FfAdd()| may be implemented as a macro. In this case, it is guaranteed
    that both arguments are evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/

/*{{{ --------------------------------------------------------------------------
!function FfSub "Finite field subtraction"
!synopsis
    FEL FfSub(FEL a, FEL b);
!arg a
    First field element.
!arg b
    Second field element.
!return
    $ |a| - |b| $
!description
    This function returns the difference of two field elements. Before you can
    use this function, the field must have been selected with |FfSetField()|.
    The arguments are not checked. If either argument is not in the allowed
    range the result is undefined and the program may crash.

    |FfSub()| may be implemented as a macro. In this case, it is guaranteed
    that both arguments are evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/

/*{{{ --------------------------------------------------------------------------
!function FfMul "Finite field multiplication"
!synopsis
    FEL FfMul(FEL a, FEL b);
!arg a
    First field element.
!arg b
    Second field element.
!return
    $ |a| \cdot |b| $
!description
    This function returns the product of two field elements. Before you can
    use this function, the field must have been selected with |FfSetField()|.
    The arguments are not checked. If either argument is not in the allowed
    range the result is undefined and the program may crash.

    |FfMul()| may be implemented as a macro. In this case, it is guaranteed
    that both arguments are evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!function FfDiv "Finite field division"
!synopsis
    FEL FfDiv(FEL a, FEL b);
!arg a
    First field element.
!arg b
    Second field element.
!return
    $ |a| / |b| $
!description
    This function returns the quotient of two field elements. Before you can
    use this function, the field must have been selected with |FfSetField()|.
    The arguments are not checked. If either argument is not in the allowed
    range, or if |b| is zero, the result is undefined and the program may
    crash.

    |FfDiv()| may be implemented as a macro. In this case, it is guaranteed
    that both arguments are evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!function FfInv "Finite field inversion"
!synopsis
    FEL FfInv(FEL a);
!arg a
    Field element.
!return
    The inverse of a.
!description
    This function returns the multiplicative inverse of a field element.
    Before you can use this function, the field must have been selected with
    |FfSetField()|. The argument is not checked. If |a| is not a valid field
    element or zero the result is undefined and the program may crash.

    |FfInv()| may be implemented as a macro. In this case, it is guaranteed
    that the argument is evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/


/*{{{ --------------------------------------------------------------------------
!function FfNeg "Additive inverse"
!synopsis
    FEL FfNeg(FEL a);
!arg a
    Field element.
!return
    $- |a| $
!description
    This function returns the additive inverse of a field element. Before
    you can use this function, the field must have been selected with
    |FfSetField()|. The argument is not checked. If |a| is not a valid field
    element the result is undefined and the program may crash.

    |FfNeg()| may be implemented as a macro. In this case, it is guaranteed
    that the argument is evaluated exactly once.
!seealso  @zarith
-------------------------------------------------------------------------- }}}*/





/* -----------------------------------------------------------------
   OpenTableFile()
   ------------------------------------------------------------------ */

static FILE *OpenTableFile(int fl)
{
    char fn[250];
    FILE *fd;

    /* Try to open the table file
       -------------------------- */
    sprintf(fn,"p%3.3d.zzz",fl);
    if ((fd = SysFopen(fn,FM_READ|FM_LIB|FM_NOERROR)) != NULL)
	return fd;

    /* Create the table file.
       ---------------------- */
    if (FfMakeTables(fl) != 0)
	MTX_ERROR("Unable to build arithmetic tables");
    fd = SysFopen(fn,FM_READ|FM_LIB);
    return fd;
}




/* -------------------------------------------------------------------------
   ReadTableFile()
   -------------------------------------------------------------------------- */

static int ReadTableFile(FILE *fd, int field)

{
    long hdr[5];

    /* Read header, perform some checks
       -------------------------------- */
    if (SysReadLong(fd,hdr,5) != 5)
    {
	MTX_ERROR("Cannot read table file header");
	return -1;
    }
    if (hdr[2] != field || hdr[1] < 0 || hdr[1] > field ||
        hdr[0] <= 1 || hdr[2] % hdr[0] != 0 || hdr[3] < 1 || hdr[3] > 8)
    {	
	MTX_ERROR("Table file is corrupted");
	return -1;
    }
    FfChar = hdr[0];
    FfGen = (FEL) hdr[1];
    MPB = hdr[3];
    if (hdr[4] != (long) ZZZVERSION)
    {
	MTX_ERROR2("Bad table file version: expected %d, found %d",
	    (int)ZZZVERSION,(int)hdr[4]);
	fclose(fd);
	return -1;
    }

    /* Read tables
       ----------- */
    if (fread(mtx_tmult,4,sizeof(mtx_tmult)/4,fd) != sizeof(mtx_tmult)/4 ||
        fread(mtx_tadd,4,sizeof(mtx_tadd)/4,fd) != sizeof(mtx_tadd)/4 ||
        fread(mtx_tffirst,1,sizeof(mtx_tffirst),fd) != sizeof(mtx_tffirst) ||
        fread(mtx_textract,1,sizeof(mtx_textract),fd) != sizeof(mtx_textract) ||
        fread(mtx_taddinv,1,sizeof(mtx_taddinv),fd) != sizeof(mtx_taddinv) ||
        fread(mtx_tmultinv,1,sizeof(mtx_tmultinv),fd) != sizeof(mtx_tmultinv) ||
        fread(mtx_tnull,1,sizeof(mtx_tnull),fd) != sizeof(mtx_tnull) ||
        fread(mtx_tinsert,1,sizeof(mtx_tinsert),fd) != sizeof(mtx_tinsert) ||
        SysReadLong(fd,mtx_embedord,4) != 4 ||
        fread(mtx_embed,16,4,fd) != 4 ||
        fread(mtx_restrict,256,4,fd) != 4
       )
    {
        MTX_ERROR("Error reading table file");
	return -1;
    }
    FfOrder = field;
    FfSetNoc(FfOrder);
    return 0;
}






/*{{{ --------------------------------------------------------------------------
!function FfSetField "Set the field"
!synopsis
    int FfSetField(int field);
!arg field
    Field order.
!return
    The function returns 0 on success, and $-1$ otherwise.
!description
    This function sets the current field to \GF{{\tt field}} and initializes
    the field arithmetic. Most kernel functions require that a field has
    been selected before they are used.

    |FfSetField()| reads an arithmetic table file into memory. If the
    file does not exist, |FfSetNoc()| tries to generate one by executing
    the |maketab| program.
!seealso FfSetNoc
-------------------------------------------------------------------------- }}}*/

int FfSetField(int field)

{
    FILE *fd;
    int result;

    if (field == FfOrder || field < 2)
	return 0;
    fd = OpenTableFile(field);
    if (fd == NULL)
    {
	MTX_ERROR1("Cannot open table file for GF(%d)",(int)field);
	return -1;
    }
    result = ReadTableFile(fd,field);
    fclose(fd);
    return result;
}



/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfSetNoc "Set the row length"
!arg ncols
    Number of columns per row.
!return
    The function returns 0 on success, and $-1$ otherwise.
!description
    This function sets the current row size to |ncols|. The row size
    is used for most low-level row operations, e.g., |FfAddRow()|.
!seealso FfSetField FfRowSize
-------------------------------------------------------------------------- }}}*/

int FfSetNoc(int ncols)

{
    MTX_ASSERT(ncols >= 0);
    FfNoc = ncols;
    if (ncols == 0)
    {
	LPR = 0;
	FfCurrentRowSize = 0;
	FfCurrentRowSizeIo = 0;
    }
    else
    {
	LPR = (ncols-1) / (MPB * sizeof(long)) + 1;	/* long's per row */
#ifdef ASM_MMX
	if (LPR % 2 != 0)
	    ++LPR;
#endif
	FfCurrentRowSize = LPR * sizeof(long);
        FfCurrentRowSizeIo = (ncols - 1) / MPB + 1;
    }
    return 0;
}


/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfRowSize "Calculate row size"
!arg noc
    Number of columns.
!return
    Row size in bytes.
!description
    This function returns the size of a row in memory, i.e., the number
    of bytes occupied by the row. Note that the row size is always a
    multiple of |sizeof(long)|. Depending on the number of columns there
    may be unused padding bytes at the end of the row.
!seealso FfTrueRowSize FfSetNoc
-------------------------------------------------------------------------- }}}*/

size_t FfRowSize(int noc)

{
    if (noc == 0)
	return 0;
    MTX_ASSERT(noc > 0);
    return ((noc-1) / (MPB * sizeof(long)) + 1) * sizeof(long);
}



/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfTrueRowSize "Number of used bytes in a row"
!arg noc
    Number of columns.
!return
    Number of bytes used by the row.
!description
    This function calculates, how many bytes are actually used by a row
    vector with |noc| columns. This number is always less or equal to
    |FfRowSize(noc)|. The `true' row size returned by this function is
    used for file i/o operations.
!seealso FfRowSize FfSetNoc
-------------------------------------------------------------------------- }}}*/

size_t FfTrueRowSize(int noc)

{
    if (noc == 0)
	return 0;
    MTX_ASSERT(noc > 0);
    return (noc-1) / MPB + 1;
}




/*{{{ --------------------------------------------------------------------------
!section kernel.ff.ff
!function FfEmbed "Embed a subfield"
!arg a
    Field element.
!arg subfield
    Subfield order.
!return
    |a|, embedded into the current field.
!description
    This function embeds an element of a subfield into the currently
    selected field. |subfield| must be a divisor of the current field order.
!seealso FfRestrict
-------------------------------------------------------------------------- }}}*/

FEL FfEmbed(FEL a, int subfield)

{   int i;

    if (subfield == FfOrder)
	return a;
    for (i = 0; mtx_embedord[i] != subfield && i < 4; ++i);
    if (i >= 4)
	MTX_ERROR2("Cannot embed GF(%d) into GF(%d)",(int)subfield,(int)FfOrder);
    return mtx_embed[i][a];
}



/*{{{ --------------------------------------------------------------------------
!function FfRestrict "Restrict to a subfield"
!arg a
    Field element.
!arg subfield
    Subfield order.
!return
    The restricted field element.
!description
    This function restricts a field element from the current field to a
    subfield. The return value
    represents the same element as |a| but with respect to the subfield.
    Note that you cannot use the return value of |FfRestrict()|
    (for field arithmetic, say) until you change to the subfield
    with |FfSetNoc(subfield,...)|.

    |a| must be of the form |FfEmbed(b,subfield)|. Otherwise
    the result is undefined. For example, restricting from \GF{8} to
    \GF{2} works only with the prime field elements, i.e., $0$ and $1$.
!seealso FfEmbed
-------------------------------------------------------------------------- }}}*/

FEL FfRestrict(FEL a, int subfield)

{
    int i;

    if (subfield == FfOrder)
	return a;
    for (i = 0; mtx_embedord[i] != subfield && i < 4; ++i);
    if (i >= 4)
    {
	MTX_ERROR2("Cannot restrict GF(%d) to GF(%d)",(int)FfOrder,
	    (int)subfield);
    }
    return mtx_restrict[i][a];
}


#ifdef ASM_MMX
static void __inline__ FastXor(void *dest, const void *src)
{
__asm__(
	
	"    pushl %ebx\n"
	"    pushl %ecx\n"
	"    pushl %edx\n"

	"    movl 8(%ebp),%ecx\n"
        "    movl 12(%ebp),%ebx\n"
        "    movl LPR,%edx\n"
        "    sarl $1,%edx\n"
        "    je .FASTXOR_1\n"
        "    .align 16\n"
	".FASTXOR_2:\n"
        "    movq (%ebx),%mm0\n"
        "    addl $8,%ebx\n"
        "    pxor (%ecx),%mm0\n"
        "    movq %mm0,(%ecx)\n"
        "    addl $8,%ecx\n"
        "    decl %edx\n"
        "    jne .FASTXOR_2\n"
	".FASTXOR_1:\n"
	"    popl %edx\n"
	"    popl %ecx\n"
	"    popl %ebx\n"
	);
}
#endif




/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfAddRow "Add two rows"
!arg dest
    The row to add to.
!arg src
    The row to add.
!return
    The function always returns |dest|.
!description
    This function adds |src| to |dest|. The row size must have been
    set with |FfSetNoc()|.
!seealso @zgaussops
-------------------------------------------------------------------------- }}}*/


    /* This assumes Intel with 4 bytes per long, but MMX implies Intel
     *        anyway.*/

PTR FfAddRow(PTR dest, PTR src)

{
    register int i;

    if (FfChar == 2)	/* characteristic 2 is simple... */
    {	
#ifdef ASM_MMX
__asm__(
	"    pushl %ebx\n"
	"    pushl %ecx\n"
	"    pushl %edx\n"

	"    movl 8(%ebp),%ecx\n"
        "    movl 12(%ebp),%ebx\n"
        "    movl LPR,%edx\n"
        "    sarl $1,%edx\n"
        "    je .ADDROW2\n"
        "    .align 16\n"
	".ADDROW1:\n"
        "    movq (%ebx),%mm0\n"
        "    addl $8,%ebx\n"
        "    pxor (%ecx),%mm0\n"
        "    movq %mm0,(%ecx)\n"
        "    addl $8,%ecx\n"
        "    decl %edx\n"
        "    jne .ADDROW1\n"
	".ADDROW2:\n"
	"    popl %edx\n"
	"    popl %ecx\n"
	"    popl %ebx\n"
	);
#else
	register long *l1 = (long *) dest;
	register long *l2 = (long *) src;
	for (i = LPR; i != 0; --i)
	{
	    register long x = *l2++;
	    if (x != 0) *l1 ^= x;
	    l1++;
	}
#endif
    }
    else		/* any other characteristic */
    {
#ifdef ASM_MMX
        register BYTE *p1 = dest;
        register unsigned long *p2 = (unsigned long *) src;
        for (i = LPR; i != 0; --i)
        {
            register unsigned long a;
            if ((a = *p2++) != 0) {
                *p1++ = mtx_tadd[*p1][a & 0xffL];
                a >>= 8;
                *p1++ = mtx_tadd[*p1][a & 0xffL];
                a >>= 8;
                *p1++ = mtx_tadd[*p1][a & 0xffL];
                a >>= 8;
                *p1++ = mtx_tadd[*p1][a & 0xffL];
            } else
              p1 += 4;
        }
#else
	register BYTE *p1 = dest;
	register BYTE *p2 = src;
	for (i = FfTrueRowSize(FfNoc); i != 0; --i)
	{
	    register int x = *p2++;
	    if (x != 0) *p1 = mtx_tadd[*p1][x];
	    p1++;
	}
#endif
    }
    return dest;
}




/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfAddRowPartial "Add a part of two rows"
!arg dest
    The row to add to.
!arg src
    The row to add.
!arg first
    Number of long int's to skip.
!arg len
    Number of long int's to add.
!return
    The function always returns |dest|.
!description
    This function works like |FfAddRow()|, but it allows the caller
    to add only a part of a row.
!seealso @zgaussops
-------------------------------------------------------------------------- }}}*/

PTR FfAddRowPartial(PTR dest, PTR src, int first, int len)
     /* first is the number of long words skipped and len the number
	of long words calculated afterwards */
{
    register long i;

    if (FfChar == 2)	/* characteristic 2 is simple... */
#ifdef ASM_MMX
__asm__("\n	movl 8(%ebp),%ecx\n"
        "	movl 12(%ebp),%ebx\n"
        "	movl 16(%ebp),%edx\n"
        "       sall $2,%edx\n"
        "       addl %edx,%ecx\n"
        "       addl %edx,%ebx\n"
        "       movl 20(%ebp),%edx\n"
        "	sarl $1,%edx\n"
        "	je .ADDROWPART_1\n"
        "	.align 16\n"
        ".ADDROWPART_2:\n"
        "	movq (%ebx),%mm0\n"
        "	addl $8,%ebx\n"
        "	pxor (%ecx),%mm0\n"
        "	movq %mm0,(%ecx)\n"
        "	addl $8,%ecx\n"
        "	decl %edx\n"
        "	jne .ADDROWPART_2\n"
        ".ADDROWPART_1:\n"
       );
#else
    {	register long *l1 = (long *) dest + first;
	register long *l2 = (long *) src + first;
	for (i = len; i != 0; --i)
	{
	    register long x = *l2++;
	    *l1 ^= x;
	    l1++;
	}
    }
#endif
    else		/* any other characteristic */
    {	register BYTE *p1 = dest + first * sizeof(long);
	register BYTE *p2 = src + first * sizeof(long);
	for (i = len*sizeof(long); i != 0; --i)
	{
	    register int x = *p2++;
	    *p1 = mtx_tadd[*p1][x];
	    p1++;
	}
    }
    return dest;
}




/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfMulRow "Multiply a row by a coefficient"
!arg row
    Pointer to the row.
!arg mark
    Coefficient.
!description
    This function multiplies |row| by the scalar |mark| using the current
    row size and returns |row|.

    Multiplying a row with zero |FF_ZERO| initializes the memory.
!seealso @zgaussops
-------------------------------------------------------------------------- }}}*/

void FfMulRow(PTR row, FEL mark)

{
    register BYTE *m;
    register BYTE *multab;
    register int i;

    CHECKFEL(mark);
    if (mark == FF_ZERO)
    {
        register long *l = (long *)row;
        for (i = LPR; i > 0; --i) *l++ = 0;
    }
    else if (mark != FF_ONE)
    {
	multab = mtx_tmult[mark];
	m = row;
	for (i = FfTrueRowSize(FfNoc); i != 0; --i)
	{
	    register int x = *m;
	    if (x != 0) *m = multab[x];
	    ++m;
	}
    }
}


/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfAddMulRow "Add a multiple of a row"
!arg dest
    The vector to add to.
!arg dest
    The vector to add.
!arg f
    Coefficient.
!description
    This function adds a multiple of |src| to |dest|.
!seealso @zgaussops
-------------------------------------------------------------------------- }}}*/

void FfAddMulRow(PTR dest, PTR src, FEL f)

{
    register int i;
    register BYTE *p1, *p2, *multab;

    CHECKFEL(f);
    if (f == FF_ZERO)
	return;
    if (f == FF_ONE)
    {
	FfAddRow(dest,src);
	return;
    }
    multab = mtx_tmult[f];
    p1 = dest;
    p2 = src;
    for (i = FfTrueRowSize(FfNoc); i != 0; --i)
    {
	*p1 = mtx_tadd[*p1][multab[*p2++]];
	++p1;
    }
}


/*{{{ --------------------------------------------------------------------------
!section kernel.ff.ff
!function FfFromInt "Convert integer to field element"
!arg l
    The integer.
!return
    The field element corresponding to |l|.
!description
    This function converts an integer to a field element using the
    same mapping as explained with |FfToInt()|. Both functions are inverse
    in the sense that the expression |f == FfFromInt(FfToInt(f))| is always
    true. |FfFromInt()| should be used whenever field elements are converted
    with |scanf()|.
!seealso FfToInt
-------------------------------------------------------------------------- }}}*/

FEL FfFromInt(int l)

{
    register int f;
    f = l % FfOrder;
    if (f < 0) f += FfOrder;
    return (FEL) f;
}






/*{{{ --------------------------------------------------------------------------
!section kernel.ff.ff
!function FfToInt "Convert field element to integer"
!arg f
    The field element.
!return
    The number corresponding to |f|.
!description
    This function converts a field element to an integer, using a `canonical'
    representation of field elements as integers which may be different from
    the internal representation. In particular, the prime field is mapped on
    $\{0,\ldots,p-1\}$. See section \ref{GdSection:kernel.ff.ff} for details.
    |FfToInt()| should be used whenever field elements are
    converted with |printf()|.
!seealso FfFromInt
-------------------------------------------------------------------------- }}}*/

int FfToInt(FEL f)

{	
    return (int) f;
}









/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfMapRow "Multiply a vector by a matrix"
!synopsis
    void FfMapRow(PTR row, PTR matrix, int nor, PTR result);
!arg row
    The vector.
!arg matrix
    The matrix.
!arg nor
    Number of rows in |matrix|.
!arg result
    Pointer to a buffer where the result will be stored.
!description
    This function multiplies the vector |row| from the right by
    the matrix |mat| and stores the result into |result|.
    |nor| is the number of rows of the matrix which must coincide
    with the size (number of columns) of the vector. The number of
    columns in both |mat| and |result| is determined by the
    current row size.

    Note: |result| and |row| must not overlap. Otherwise the result is
    undefined.
!seealso
-------------------------------------------------------------------------- }}}*/

void FfMapRow(PTR row, PTR matrix, int nor, PTR result)

{
    register int i;
    register FEL f;
    BYTE *m = (BYTE *) matrix;
    register long *l = (long *)result;

#ifdef DEBUG
    if (result >= row && result < row + FfRowSize(nor))
	MTX_ERROR("row and result overlap: undefined result!");
    if (row >= result && row < result + FfCurrentRowSize)
	MTX_ERROR("row and result overlap: undefined result!");
#endif

    /* Fill the result with zeroes.
       ---------------------------- */
    for (i = LPR; i > 0; --i)
	*l++ = 0;

    if (FfOrder == 2)       /* GF(2) is a special case */
    {
        register long *x1 = (long *) matrix;
        register BYTE *r = (BYTE *) row;

        for (i = nor; i > 0; ++r)
        {
	    register BYTE mask;
	    if (*r == 0)
	    {
		i -= 8;
		x1 += 8 * LPR;
		continue;
	    }
	    for (mask = 0x80; mask != 0 && i > 0; mask >>= 1, --i)
	    {
                if ((mask & *r) == 0)
		{
                    x1 += LPR;  /* Skip that row */
		    continue;
		}
#ifdef ASM_MMX
__asm__("    pushl %ebx\n");
__asm__("    movl %0,%%ebx" : : "g" (x1) );
__asm__("    pushl %ecx\n"
	"    pushl %edx\n"
	"    movl 20(%ebp),%ecx\n"	/* result */
	);
__asm__ (
        "    movl LPR,%edx\n"
        "    sarl $1,%edx\n"
        "    je .FASTXOR_1\n"
        "    .align 16\n"
	".FASTXOR_2:\n"
        "    movq (%ebx),%mm0\n"
        "    addl $8,%ebx\n"
        "    pxor (%ecx),%mm0\n"
        "    movq %mm0,(%ecx)\n"
        "    addl $8,%ecx\n"
        "    decl %edx\n"
        "    jne .FASTXOR_2\n"
	".FASTXOR_1:\n"
	"    popl %edx\n"
	"    popl %ecx\n");
__asm__("    movl %%ebx,%0" : : "g" (x1) );
__asm__("    popl %ebx\n"
	);

#else
		{
                register long *x2 = (long *)result;
                register int k;
                for (k = LPR; k; --k)
                    *x2++ ^= *x1++;
		}
#endif
            }
        }
    }
    else                /* Any other field */
    {
        register BYTE *brow = (BYTE *) row;
        register int pos = 0;

        for (i = nor; i > 0; --i)
        {
            f = mtx_textract[pos][*brow];
            if (++pos == (int) MPB)
            {
                pos = 0;
                ++brow;
            }
            if (f != FF_ZERO)
            {
                register BYTE *v = m;
                register BYTE *r = result;
                register int k = FfCurrentRowSizeIo;
                if (f == FF_ONE)
                {
                    for (; k != 0; --k)
                    {
                        *r = mtx_tadd[*r][*v++];
                        ++r;
                    }
                }
                else
                {
                    register BYTE *multab = mtx_tmult[f];
                    for (; k != 0; --k)
                    {
		 	if (*v != 0)
			    *r = mtx_tadd[multab[*v]][*r];
			++v;
                        ++r;
                    }
                }
            }
            m += FfCurrentRowSize;              /* next row */
        }
    }
}






/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfScalarProduct "Scalar product of two vectors"
!arg a
    The first vector.
!arg b
    The second vector.
!return
    Scalar product of the two vectors.
!description
    Given two vectors $a=(a_i)$ and $b=(b_i)$, this function calculates the
    scalar product $p=\sum_ia_ib_i$.
!seealso
-------------------------------------------------------------------------- }}}*/


FEL FfScalarProduct(PTR a, PTR b)

{
    register unsigned char *ap = (unsigned char *) a;
    register unsigned char *bp = (unsigned char *) b;
    register int i;
    FEL f = FF_ZERO;

    for (i = FfNoc-1; i >= MPB; i -= MPB)
    {

	register int k;
	if (*ap != 0 && *bp != 0)
	{
	    for (k = 0; k < MPB; ++k)
		f = FfAdd(f,FfMul(mtx_textract[k][*ap],mtx_textract[k][*bp]));
	}
	++ap;
	++bp;
    }
    for (; i >= 0; --i)
	f = FfAdd(f,FfMul(mtx_textract[i][*ap],mtx_textract[i][*bp]));

    return f;
}




/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfExtractColumn "Extract one column of a matrix"
!arg mat
    Pointer to the matrix.
!arg nor
    Number of rows in matrix.
!arg col
    Column to extract (starting with 1).
!arg result
    Pointer to buffer for the extracted column.
!description
    This function extracts one column out of a matrix and stores it as
    a row vector in |result|. The number of columns of the matrix must have
    been set with |FfSetNoc()|. |nor| is the number of rows in the matrix. The
    result is a row with |nor| entries, i.e., the length of |result| must be at
    least |nor|. |mat| and |result| must not overlap, or the result is
    undefined.
-------------------------------------------------------------------------- }}}*/

void FfExtractColumn(PTR mat,int nor,int col,PTR result)

{
    register BYTE *x = (BYTE *)mat + (col / MPB);
    register BYTE *extab = mtx_textract[col % MPB];
    register BYTE a = 0;
    register int ind = 0;
    register BYTE *y = result;
    register int count;

    for (count = nor; count > 0; --count)
    {
	a = (BYTE) (a + mtx_tinsert[ind][extab[*x]]);
	if (++ind == MPB)
	{
	    *y++ = a;
	    a = 0;
	    ind = 0;
	}
	x += FfCurrentRowSize;
    }
    if (ind != 0) *y = a;
}


























/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfInsert  "Insert a mark into a row"
!synopsis
    void FfInsert(PTR row, int col, FEL mark);
!arg row
    Pointer to the row.
!arg col
    Insert position (0-based).
!arg mark
    Value to insert.
!description
    |FfInsert()| inserts the field element |mark| at position
    |col| into |row|. {\em Note}: |col| is a 0-based index.
    Thus, to insert a value into the first position of a row, use
    |FfInsert(row,0,value)|. Before this function can be used, a field
    must be selected with |FfSetNoc()|.

    |FfInsert()| does not need the row size beeing set correctly. For
    example, if you are working with rows of different size, you do
    not have to call |FfSetNoc()| prior to each |FfInsert()|. On the
    other hand, there is no protection against writing beyond the end
    of a row.

    If the MeatAxe is compiled with the DEBUG option (see section
    \ref{sec:compilation}, |FfInsert()| checks that |mark| is a valid
    field element and |col| is not negative. If also the PARANOID option
    was in effect during compilation, |FfInsert()| checks if |col| is
    less than or equal to the current row size.
!seealso FfExtract
-------------------------------------------------------------------------- }}}*/

void FfInsert(PTR row, int col, FEL mark)

{
    register BYTE *loc = (BYTE *)row + (col / MPB);
    register int idx = col % MPB;

#ifdef DEBUG
    CHECKFEL(mark);
    if (col < 0)
	MTX_ERROR1("Invalid column index %d",col);
#ifdef PARANOID
    if (col >= FfNoc)
	MTX_ERROR2("Invalid column index %d, noc=%d)",col,FfNoc);
#endif
#endif
    *loc = (BYTE) (mtx_tnull[idx][*loc] + mtx_tinsert[idx][mark]);
}



/*{{{ --------------------------------------------------------------------------
!function FfExtract  "Extract a mark from a row"
!arg row
    Pointer to the row.
!arg col
    Index of mark to extract (0-based).
!return
    |col|-th entry of |row|.
!description
    This function returns the entry at position |col| of a
    row.  Note that column indexes start with 0, i.e., |FfExtract(row,0)|
    returns the first entry of a row. Like |FfInsert()|, this function
    does not depend on the current row size. Reading beyond the end of
    a row will probably not produce an error, but the result is undefined.
!seealso FfInsert
-------------------------------------------------------------------------- }}}*/

FEL FfExtract(PTR row, int col)

{
#ifdef DEBUG
    if (col < 0)
	MTX_ERROR1("Invalid column index (%d)",col);
#ifdef PARANOID
    if (col >= FfNoc)
	MTX_ERROR2("Invalid column index %d, noc=%d",col,FfNoc);
    CHECKFEL(mtx_textract[col % MPB][((BYTE *)row)[col / MPB]]);
#endif
#endif
    return mtx_textract[col % MPB][((BYTE *)row)[col / MPB]];
}



/*{{{ --------------------------------------------------------------------------
!function FfFindPivot "Find pivot column"
!arg row
    Pointer to the row.
!arg mark
    Buffer for pivot element.
!return
    Index of the first non-zero entry in |row|, or $-1$ if all entries
    are zero.
!description
    This function scans the vector |row| and finds the first non-zero
    mark. The mark is stored into |*mark| and its position (counting
    from 0) is returned. If the whole vector is zero, |FfFindPivot()|
    returns $-1$ and leaves |*mark| unchanged.
!seealso
-------------------------------------------------------------------------- }}}*/

int FfFindPivot(PTR row, FEL *mark)

{
    register long *l = (long *) row;
    register int idx;
    register BYTE *m;

    for (idx = 0; idx < LPR && *l == 0; ++idx, ++l);
    if (idx == LPR)
	return -1;
    idx = idx * sizeof(long) * MPB;
    m = (BYTE *)l;
    while (*m == 0)
    {	++m;
	idx += MPB;
    }
    idx += mtx_tffirst[*m][1];
    if (idx >= FfNoc)		/* Ignore garbage in padding bytes */
	return -1;
    *mark = mtx_tffirst[*m][0];
    return idx;
}



