/* ============================= C MeatAxe ==================================
   File:        $Id: matcut.c,v 1.1.1.1 2000/01/07 16:42:26 mringe Exp $
   Comment:     Cut a rectangular piece out of a matrix.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"

   
/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO




/*{{{ --------------------------------------------------------------------------
!section obj.mat
!function MatCut "Cut a rectangle out of a matrix"
!arg src
    Pointer to the matrix.
!arg row1
    First row in region.
!arg col1
    First column in region.
!arg nrows
    Number of rows to cut. $-1$ means as many rows as possible.
!arg ncols
    Number of columns to cut. $-1$ means as many columns as possible.
!return
    Pointer to a new matrix containing the specified region of |src|, or 
    |NULL| on error.
!description
    This function cuts a rectangular region out of a matrix. |row1| and |col1|
    define the upper left corner of the region. As usual, row and column indexes 
    start with 0. |nrows| and |ncols| are the height (number of rows) and width 
    (number of columns) of the region. Both |nrows| and |ncols| can be $-1$.
    In this case the region extends up to the last row or last column, 
    respectivly. For example, to extract the first 10 rows from a matrix 
    independently of the number of columns, you could say 
    |MatCut(mat,0,0,10,-1)|.
    The rectangle defined by |row1|, |col1|, |nrows| and |ncols| must not
    exceed the size of the matrix.

    The function returns a new matrix containing a copy of the specified
    rectangle. In case of error, the return value is |NULL|.
!seealso MatCopyRegion MatCutRows
-------------------------------------------------------------------------- }}}*/

Matrix_t *MatCut(const Matrix_t *src, int row1, int col1, int nrows, int ncols)

{
    Matrix_t *result;
    PTR s, d;
    int n;
    
    /* Check arguments
       --------------- */
#ifdef DEBUG
    if (!MatIsValid(src))
	return NULL;
#endif
    if (nrows == -1)
	nrows = src->Nor - row1;
    if (ncols == -1)
	ncols = src->Noc - col1;
    if (row1 < 0 || nrows < 0 || row1 + nrows > src->Nor)
    {
	MTX_ERROR("Source row index out of bounds");
	return NULL;
    }
    if (col1 < 0 || ncols < 0 || col1 + ncols > src->Noc)
	return MTX_ERROR("Source column index out of bounds"), NULL;

    /* Allocate a new matrix for the result
       ------------------------------------ */
    result = MatAlloc(src->Field,nrows,ncols);
    if (result == NULL)
	return NULL;
    if (nrows == 0)
	return result;

    /* Initialize pointers to the source and destination matrix 
       -------------------------------------------------------- */
    s = MatGetPtr(src,row1);
    d = result->Data;

    /* Copy the requested data
       ----------------------- */
    FfSetNoc(ncols);
    for (n = nrows; n > 0; --n)
    {
	if (col1 == 0)
	    FfCopyRow(d,s);
	else
	{
	    register int k;
	    for (k = 0; k < ncols; ++k)
	    {
#ifdef PARANOID
		FEL f;
		FfSetNoc(src->Noc);
		f = FfExtract(s,col1+k);
		FfSetNoc(ncols);
		FfInsert(d,k,f);
#else
		FfInsert(d,k,FfExtract(s,col1+k));
#endif
	    }
	}
	FfStepPtr(&d);
	s = (PTR)((char *)s + src->RowSize);
    }

    return result;
}




/*{{{ --------------------------------------------------------------------------
!function MatCutRows "Cut rows out of a matrix"
!arg src
    Pointer to the matrix.
!arg row1
    First row in region.
!arg nrows
    Number of rows to cut. $-1$ means as many rows as possible.
!return
    Pointer to a new matrix containing the specified rows of |src|, or 
    |NULL| on error.
!description
    This function cuts a range of rows, specified by |row1| and |nrows| out of
    a matrix. |row1| is the first row of the range (starting with 0), and |nrows|
    is the number of rows to cut. |nrows| may be $-1$, meaning `as many rows as
    possible'. |row1| and |nrows| must specify a valid range, i.e., |row1| must
    be a valid row index, |nrows| must be greater or equal to $-1$, and, if 
    |nrows| is not $-1$, $|row|+|nrows|$ must be less than the number of rows 
    of the matrix.

    The function returns a new matrix containing a copy of the specified
    rows. In case of error, the return value is |NULL|.
!seealso MatCopyRegion MatCutRows
-------------------------------------------------------------------------- }}}*/

Matrix_t *MatCutRows(const Matrix_t *src, int row1, int nrows)

{
    return MatCut(src,row1,0,nrows,-1);
}




