/* ============================= C MeatAxe ==================================
   File:        $Id: matins.c,v 1.1.1.1 2000/01/07 16:42:26 mringe Exp $
   Comment:     Insert a matrix into a polynomial
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"

MTX_DEFINE_FILE_INFO



/*{{{ --------------------------------------------------------------------------
!section algo.etc
!function MatInsert_ "Insert a matrix into a polynomial"
!synopsis 
    Matrix_t *MatInsert_(Matrix_t *mat, const Poly_t *pol);
!arg mat
    Pointer to the matrix.
!arg pol
    Pointer to the polynomial.
!return
    The function returns |mat|, or |NULL| on error.
!description
    This function inserts a matrix $M$ into a polynomial $p(x)$. This function 
    is destructive. It stores $p(M)$ in |mat|, and does not allocate a new 
    matrix for the result.
    |mat| must be a square matrix, and |pol| must be over the same field 
    as |mat|.
!seealso @matinsert
-------------------------------------------------------------------------- }}}*/

Matrix_t *MatInsert_(Matrix_t *mat, const Poly_t *pol)

{
    Matrix_t *x = NULL;
    int i;
    int nor;
    int l;
    PTR v;
    FEL f;

    /* Check the arguments
       ------------------- */
    if (!MatIsValid(mat))
	return NULL;
    if (!PolIsValid(pol))
	return NULL;
    if ((nor = mat->Nor) != mat->Noc) 
    {
	MTX_ERROR1("%E",MTX_ERR_NOTSQUARE);
	return NULL;
    }
    if (mat->Field != pol->Field) 
    {
	MTX_ERROR1("%E",MTX_ERR_INCOMPAT);
	return NULL;
    }

    FfSetField(mat->Field);
    FfSetNoc(nor);

    /* Special case: p(x) = 0
       ---------------------- */
    if (pol->Degree == -1)
    {
	for (l = 0, v = mat->Data; l < nor; FfStepPtr(&v), ++l)
	    FfMulRow(v,FF_ZERO);
	return mat;
    }

    /* Special case: deg(p) = 0
       ------------------------ */
    if (pol->Degree == 0)
    {
	for (l = 0, v = mat->Data; l < nor; FfStepPtr(&v), ++l)
	{
	    FfMulRow(v,FF_ZERO);
	    FfInsert(v,l,pol->Data[0]);
	}
	return mat;
    }

    /* Evaluate p(A)
       ------------- */
    if (pol->Degree > 1) 
	x = MatDup(mat);
    if ((f = pol->Data[pol->Degree]) != FF_ONE)
    {
	for (l = nor, v = mat->Data; l > 0; --l, FfStepPtr(&v))
	    FfMulRow(v,f);
    }
    for (i = pol->Degree - 1; i >= 0; --i)
    {
        if ((f = pol->Data[i]) != FF_ZERO)
        {
	    for (l = 0, v = mat->Data; l < nor; ++l, FfStepPtr(&v))
		FfInsert(v,l,FfAdd(FfExtract(v,l),f));
        }
	if (i > 0) 
	    MatMul(mat,x);
    }
    if (pol->Degree > 1) MatFree(x);
    return mat;
}



/*{{{ --------------------------------------------------------------------------
!function MatInsert "Insert a matrix into a polynomial"
!synopsis 
    Matrix_t *MatInsert(const Matrix_t *mat, const Poly_t *pol);
!arg mat
    Pointer to the matrix.
!arg pol
    Pointer to the polynomial.
!return
    The function returns |mat|, or |NULL| on error.
!description
    This function inserts a matrix into a polynomial. Unlike |MatInsert_()|,
    this function is not destructive. It allocates a new matrix for the result
    and does not modify the original matrix.
    |mat| must be a square matrix, and |pol| must be over the same field 
    as |mat|.
!seealso @matinsert
-------------------------------------------------------------------------- }}}*/

Matrix_t *MatInsert(const Matrix_t *mat, const Poly_t *pol)

{
    Matrix_t *x;
    int i;
    int nor;
    int l;
    PTR v;
    FEL f;

    if (!MatIsValid(mat))
	return NULL;
    if (!PolIsValid(pol))
	return NULL;
    if ((nor = mat->Nor) != mat->Noc) 
    {
	MTX_ERROR1("%E",MTX_ERR_NOTSQUARE);
	return NULL;
    }
    if (mat->Field != pol->Field) 
    {
	MTX_ERROR1("%E",MTX_ERR_INCOMPAT);
	return NULL;
    }

    /* Special case: p = 0
       ------------------- */
    if (pol->Degree == -1)
	return MatAlloc(mat->Field,nor,nor);

    /* Special case: deg(p) = 0
       ------------------------ */
    if (pol->Degree == 0)
    {
	x = MatAlloc(mat->Field,nor,nor);
	for (l = 0, v = x->Data; l < nor; ++l, FfStepPtr(&v))
	    FfInsert(v,l,pol->Data[0]);
	return x;
    }

    /* Evaluate p(A)
       ------------- */
    x = MatDup(mat);
    if ((f = pol->Data[pol->Degree]) != FF_ONE)
    {
	for (l = nor, v = x->Data; l > 0; --l, FfStepPtr(&v))
	    FfMulRow(v,f);
    }
    for (i = pol->Degree - 1; i >= 0; --i)
    {
        if ((f = pol->Data[i]) != FF_ZERO)
        {
	    for (l = 0, v = x->Data; l < nor; ++l, FfStepPtr(&v))
		FfInsert(v,l,FfAdd(FfExtract(v,l),f));
        }
	if (i > 0) 
	    MatMul(x,mat);
    }
    return x;
}

