#include "meataxe.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

MTX_DEFINE_FILE_INFO

static MtxApplicationInfo_t AppInfo = { 
"mtxsh", "MeatAxe shell", 
"\n"
"SYNTAX\n"
"    mtxsh [-e <Command>] [-c <Script] [<Arg> ...]\n"
"\n"
"OPTIONS\n"
"    -e ...................... Execute <Command>\n"
"    -f ...................... Read commands from <File>\n"
"\n"
"COMMANDS\n"
"    add ..................... Add matrices\n"
"    clean ................... Clean a matrix by a space\n"
"    make-word ............... Make word\n"
"    mul ..................... Multiply matrices or permutations\n"
"    null-space .............. Matrix null-space\n"
"    power ................... Power of a matrix or permutation\n"
"    random-orders ........... Random orders\n"
};

static MtxApplication_t *App = NULL;
static int HaveScript = 0;



typedef struct 
{
    const char *ShortCut;
    const char *Name;
    int (*Function)(int, const char **);
} FuncTab_t;

static FuncTab_t FuncTab[] =
{
    {"zad","add", YAdd },
    {"zcf","change-format", YChangeFormat },
    {"zcl","clean", YClean },
    {"zcp","charpol", YCharPol },
    {"zef","echelon", YEchelon },
    {"zmu","mul", YMul },
    {"zmw","make-word", YMakeWord },
    {"znu","null-space", YNullspace },
    {"zpo","power", YPower },
    {"zro","random-orders", YRandomOrders },
    { NULL, NULL, NULL }
};




// -----------------------------------------------------------------
// Command execution

static int Execute(int argc, const char **argv)

{
    const char *command;
    const FuncTab_t *ft;

    if (argc < 1)
	return 0;
    command = argv[0];
    for (ft = FuncTab; ft->Function != NULL; ++ft)
    {
	if (!strcmp(ft->Name,command))
	    return ft->Function(argc,argv);
    }
    MTX_ERROR1("Unknown command '%s'",command);
    return -1;
}


// -----------------------------------------------------------------
// Script management

typedef struct
{
    int LineNo;
    const char *Text;
} ScriptLine_t;



static int ScriptSize = 0;
static ScriptLine_t **Script = NULL;


static char *StrDup(const char *c)

{
    char *copy = malloc(strlen(c)+1);
    if (copy == NULL)
	return NULL;
    strcpy(copy,c);
    return copy;
}


static int AddScriptLine(int lineno, const char *text)

{
    ScriptLine_t *sl;
    ScriptLine_t **new_script;

    HaveScript = 1;
    new_script = NREALLOC(Script,ScriptLine_t *,ScriptSize + 1);
    if (new_script == NULL)
	return -1;
    Script = new_script;
    sl = Script[ScriptSize++] = ALLOC(ScriptLine_t);
    if (sl == NULL)
	return -1;
    sl->LineNo = lineno;
    sl->Text = StrDup(text);
    if (sl->Text == NULL)
	return -1;
    return 0;
}




static int AddScriptFile(const char *file_name)

{
    char buf[4000];
    FILE *f;
    int lineno = 0;

    if ((f = SysFopen(file_name,FM_READ|FM_TEXT)) == NULL)
	return -1;
    while (!feof(f) && fgets(buf,sizeof(buf),f) != NULL)
    {
	AddScriptLine(++lineno,buf);
	buf[0] = 0;
    }
    fclose(f);
}





static void CheckShortcut(int argc, const char **argv)

{
    const FuncTab_t *ft;
    const char *c;
    for (c = argv[0]; *c != 0; ++c)
	;
    while (c != argv[0] && strchr("/\\:",c[-1] == 0))
	--c;
    for (ft = FuncTab; ft->Function != NULL; ++ft)
    {
	if (!strcmp(ft->ShortCut,c))
	{
	    int rc;
	    rc = ft->Function(argc,argv);
	    exit(rc);
	}
    }
}




static int Init(int argc, const char **argv)

{
    const char *c;
    int i = 0;


    CheckShortcut(argc,argv);

    if ((App = AppAlloc(&AppInfo,argc,argv)) == NULL)
	return -1;

    /* Command line options.
       --------------------- */
    while ((c = AppGetTextOption(App,"-c",NULL)) != NULL)
    {
	if (AddScriptLine(++i,c) != 0)
	    return -1;
    }
    while ((c = AppGetTextOption(App,"-f",NULL)) != NULL)
    {
	if (AddScriptFile(c) != 0)
	    return -1;
    }

    if (AppGetArguments(App,0,100) < 0)
	return -1;

    return 0;
}



static void Cleanup()

{
    AppFree(App);
}



// -----------------------------------------------------------------
// Script execution

#define SCRIPT_MAX_ARGC 50
int ArgC;
const char *ArgV[SCRIPT_MAX_ARGC+1];


static int SplitCmd(char *c)

{
    ArgC = 0;
    ArgV[0] = strtok(c," \t\n\r");
    while (ArgV[ArgC] != NULL)
    {
	if (ArgC >= SCRIPT_MAX_ARGC)
	{
	    MTX_ERROR("Too many arguments");
	    return -1;
	}
	ArgV[++ArgC] = strtok(NULL," \t\n\r");
    }
    return 0;
}


static int IsEmpty(const char *c)

{
    while (*c != 0 && isspace(*c))
	++c;
    return (*c == 0 || *c == '#');
}



static int Expand(const char *c, char *buf, int buf_size)

{
    char *d = buf;
    int free_space = buf_size;

    while (*c != 0 && free_space > 1)
    {
	if (*c == '$' && isdigit(c[1]))
	{
	    const char *a;
	    int i = atoi(c+1);
	    if (i < 1 || i > App->ArgC)
	    {
		MTX_ERROR1("Invalid argument $%d",i);
		return -1;
	    }
	    a = App->ArgV[i-1];
	    while (*a != 0 && free_space > 1)
	    {
		*d++ = *a++;
		--free_space;
	    }
	    /* Skip the number */
	    do ++c; while (*c != 0 && isdigit(*c));
	}
	else
	{
	    *d++ = *c++;
	    free_space--;
	}
    }
    if (free_space <= 0)
    {
	MTX_ERROR("Command too complex");
	return -1;
    }
    *d = 0;

    return 0;
}



static int ExecLine(const char *c)

{
    static char cmd_buffer[2000];

    if (IsEmpty(c))
	return 0;
    if (Expand(c,cmd_buffer,sizeof(cmd_buffer)) != 0)
	return -1;
    if (SplitCmd(cmd_buffer) != 0)
	return -1;
    return Execute(ArgC,ArgV);
}


static int RunScript()

{ 
    int i;
    for (i = 0; i < ScriptSize; ++i)
    {
	if (ExecLine(Script[i]->Text) != 0)
	{
	    MTX_ERROR1("Error at line %d",Script[i]->LineNo);
	    return -1;
	}
    }
    return 0;
}


static int Loop()

{
    char buf[2000];

    while (!feof(stdin))
    {
	printf("mtxsh>");
	fflush(stdout);
	gets(buf);
	if (ExecLine(buf) != 0)
	    return -1;
    }
    return 0;
}



// -----------------------------------------------------------------
// Main


int main(int argc, char *argv[])

{
    int result;

    if (Init(argc,(const char **)argv) != 0)
	return 1;

    if (HaveScript)
	result = RunScript();
    else
	result = Loop();

    Cleanup();
    return 0;
}