/* ============================= C MeatAxe ==================================
   File:        $Id: polcore.c,v 1.1.1.1 2000/01/07 16:42:27 mringe Exp $
   Comment:     Basic polynomial functions.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>

   
/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO

#define POLY_MAGIC 0x355A3207



/*{{{ --------------------------------------------------------------------------
!section obj.poly
!function PolIsValid "Check a polynomial"
!arg p
    Pointer to the polynomial.
!return
    1 if |p| points to a valid polynomial, 0 otherwise.
!description
    This function checks if the argument, |p|, is a pointer to a valid
    polynomial. If the polynomial is o.k., the function returns 1. 
    Otherwise, an error is signaled and, if the error handler does not 
    terminate the program, the function returns 0.
!seealso Poly_t
-------------------------------------------------------------------------- }}}*/

int PolIsValid(const Poly_t *p)

{
    int deg;

    if (p == NULL)
    {
	MTX_ERROR("NULL polynomial");
	return 0;
    }
    deg = p->Degree;
    if (p->Magic != POLY_MAGIC || deg < -1 || p->Field < 2 
	|| p->Data == NULL || p->BufSize < 0)
    {
	MTX_ERROR4("Invalid polynomial (magic=%d, field=%d, deg=%d, bufsize=%d)",
	    p->Magic,p->Field,deg,p->BufSize);
	return 0;
    }

    if (deg >= 0 && p->Data[deg] == FF_ZERO)
    {
	MTX_ERROR("Invalid polynomial: leading coefficient is zero");
	return 0;
    }
    return 1;
}







/*{{{ --------------------------------------------------------------------------
!function PolAlloc "Allocate a polynomial"
!arg fl
    Field order.
!arg n
    Degree of the polynomial.
!return
    Pointer to a new |Poly_t| structure or |NULL| on error.
!description
    This function creates the polynomial $p(x)=x^n$ over the current field.
    If |n| is negative, a zero polynomial is created. The return value is a 
    pointer to a newly allocated |Poly_t| structure. The caller is responsible 
    for releasing memory by calling |PolFree()| when the polynomial is no 
    longer needed.
!seealso PolFree
-------------------------------------------------------------------------- }}}*/

Poly_t *PolAlloc(int fl, int n)

{
    Poly_t *x;
    int i, s;

    if (n < 0) 
	n = -1;
    if ((s = n + 1) < 1) 
	s = 1;

    FfSetField(fl);
    if ((x = ALLOC(Poly_t)) == NULL) 
    {
	MTX_ERROR("Cannot allocate polynomial");
	return NULL;
    }
    x->Magic = POLY_MAGIC;
    x->Field = fl;
    x->Degree = n;
    x->BufSize = s;
    if ((x->Data = NALLOC(FEL,s)) == NULL)
    {
	SysFree(x);
        MTX_ERROR("Cannot allocate polynomial data");
        return NULL;
    }
    for (i = 0; i < (int) s-1; ++i) 
	x->Data[i] = FF_ZERO;
    x->Data[s-1] = FF_ONE;
    return x;
}




/*{{{ --------------------------------------------------------------------------
!function PolFree "Free a polynomial"
!section obj.poly
!synopsis 
    int PolFree(Poly_t *x);
!arg x
    Pointer to the polynomial.
!return
    $0$ on success, $-1$ on error.
!description
    This function frees a polynomial data structure and cleans up all internal
    data.
!seealso PolAlloc
-------------------------------------------------------------------------- }}}*/

int PolFree(Poly_t *x)

{
    if (!PolIsValid(x))
	return -1;
    SysFree(x->Data);
    memset(x,0,sizeof(Poly_t));
    SysFree(x);
    return 0;
}



/* ------------------------------------------------------------------
   Pol_Normalize()
   ------------------------------------------------------------------ */

void Pol_Normalize(Poly_t *p)

{
    int i = p->Degree;

    while (i >= 0 && p->Data[i] == FF_ZERO) 
	--i;
    p->Degree = i;
}



