/* ============================= C MeatAxe ==================================
   File:        $Id: poldiv.c,v 1.1.1.1 2000/01/07 16:42:27 mringe Exp $
   Comment:     Polynomial division.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>

   
/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO


/*{{{ --------------------------------------------------------------------------
!section obj.poly
!function PolDivMod "Polynomial division"
!arg a
    Pointer to the first polynomial (numerator).
!arg b
    Pointer to the second polynomial (denominator).
!return
    The function returns the quotient, or |NULL| on error.
!description
    This function performs a polynomial division. Given two polynomials $a$ and
    $b\neq 0$, which must be over the same field, |PolDivMod()| finds two 
    polynomials $q$ and $r$ such that $a=qb+r$, and $\deg(r)<\deg(b)$.

    The quotient, $q$, is returned as the function result. This is a newly 
    allocated polynomial. The caller is responsible for deleting the quotient
    when it no longer needed.

    The remainder, $r$, is stored in |a| and replaces the original value. If you
    need to keep the value of |a| you must make a copy using |PulDup()| before
    calling |PolDivMod()|. |b| is not changed.
!seealso @polymod
-------------------------------------------------------------------------- }}}*/

Poly_t *PolDivMod(Poly_t *a, const Poly_t *b)

{
    Poly_t *q;

    /* Check arguments
       --------------- */
    if (!PolIsValid(a) || !PolIsValid(b))
	return NULL;
    if (a->Field != b->Field) 
    {
	MTX_ERROR1("%E",MTX_ERR_INCOMPAT);
	return NULL;
    }
    FfSetField(a->Field);
    if (b->Degree <= -1) 
    {
	MTX_ERROR1("%E",MTX_ERR_DIV0);
	return NULL;
    }
    if (a->Degree < b->Degree)
    {
	q = PolAlloc(a->Field,-1);	/* Trivial case: Quotient = 0 */
    }
    else
    {
	FEL lead = b->Data[b->Degree];
	int i, k;

	if (lead == FF_ZERO) 
	{
	    MTX_ERROR1("%E",MTX_ERR_DIV0);
	    return NULL;
	}
	q = PolAlloc(FfOrder,a->Degree - b->Degree);
	if (q == NULL) 
	{
	    MTX_ERROR("Cannot allocate result");
	    return NULL;
	}
	for (i = a->Degree; i >= b->Degree; --i)
	{
	    FEL qq = FfNeg(FfDiv(a->Data[i],lead));
	    for (k = 0; k <= b->Degree; ++k)
		a->Data[i-k] = FfAdd(a->Data[i-k],
		    FfMul(qq,b->Data[b->Degree - k]));
	    q->Data[i-b->Degree] = FfNeg(qq);
	}
        Pol_Normalize(a);
    }
    return q;
}




/*{{{ --------------------------------------------------------------------------
!function PolMod "Polynomial division"
!arg a
    Pointer to the first polynomial (numerator).
!arg b
    Pointer to the second polynomial (denominator).
!return
    The function returns |a|, which has been reduced modulo |b|. On error
    the function returns |NULL|.
!description
    This function reduces |a| modulo |b|.
!seealso @polymod
-------------------------------------------------------------------------- }}}*/

Poly_t *PolMod(Poly_t *a, const Poly_t *b)

{

    /* Check arguments
       --------------- */
    if (!PolIsValid(a) || !PolIsValid(b))
	return NULL;
    if (a->Field != b->Field) 
    {
	MTX_ERROR1("%E",MTX_ERR_INCOMPAT);
	return NULL;
    }

    FfSetField(a->Field);
    if (b->Degree <= -1) 
    {
	MTX_ERROR1("%E",MTX_ERR_DIV0);
	return NULL;
    }
    if (a->Degree >= b->Degree)
    {
	FEL lead = b->Data[b->Degree];
	int i, k;

	if (lead == FF_ZERO) 
	{
	    MTX_ERROR1("%E",MTX_ERR_DIV0);
	    return NULL;
	}
	for (i = a->Degree; i >= b->Degree; --i)
	{
	    FEL qq = FfNeg(FfDiv(a->Data[i],lead));
	    for (k = 0; k <= b->Degree; ++k)
		a->Data[i-k] = FfAdd(a->Data[i-k],
		    FfMul(qq,b->Data[b->Degree - k]));
	}
        Pol_Normalize(a);
    }
    return a;
}





