/* ============================= C MeatAxe ==================================
   File:        $Id: quotient.c,v 1.1.1.1 2000/01/07 16:42:27 mringe Exp $
   Comment:     Projection on quotient.
   --------------------------------------------------------------------------
   (C) Copyright 1997 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <string.h>
#include <stdlib.h>






/* --------------------------------------------------------------------------
   Local data
   -------------------------------------------------------------------------- */

MTX_DEFINE_FILE_INFO 




/*{{{ --------------------------------------------------------------------------
!section algo.split
!function QProjection "Projection on quotient"
!synopsis 
    Matrix_t *QProjection(const Matrix_t *subspace, const Matrix_t *vectors);
!return
    Projection of |vectors| on the quotient by |subspace|, or |NULL| on error.
!description
    This function calculates the projection of a matrix onto the quotient by a 
    subspace. The first matrix, |subspace| must be in echelon form, while the
    second argument can be any matrix. Of course both matrices must be over the
    same field and have the same number of columns. The return value is a 
    pointer to a matrix containing the projections the |vectors|. This matrix 
    is not in echelon form an may even contain null rows.

    The projection depends on the basis for the subspace and is calculated as
    follows. Let $V=F^{n\times n}$ and $w_1,\ldots,w_s$ be a basis 
    for the subspace $W\leq V$. The basis, written as a matrix of row vectors, 
    is assumed to be in semi-echelon form. By looking at the pivot columns we 
    can construct the vectors $w_{s+1},\ldots,w_n$ by taking all vectors which
    have a exactly one 1 at any non-pivot position and are zero otherwise.
    Then, $(w_1,\ldots,w_s,w_{s+1},\ldots,w_n)$ is a basis for $V$ in
    semi-echelon form and defines the decomposition of any vector into
    subspace and quotient part.
!seealso 
-------------------------------------------------------------------------- }}}*/

Matrix_t *QProjection(const Matrix_t *subspace, const Matrix_t *vectors)

{
    int i, sdim, qdim;
    int *non_piv;
    Matrix_t *result;
    PTR tmp;

    /* Check the arguments
       ------------------- */
    if (!MatIsValid(subspace) || !MatIsValid(vectors))
	return NULL;
    if (subspace->Field != vectors->Field || subspace->Noc != vectors->Noc)
    {
	MTX_ERROR1("%E",MTX_ERR_INCOMPAT);
	return NULL;
    }
    if (subspace->PivotTable == NULL)
    {
	MTX_ERROR1("%E",MTX_ERR_NOTECH);
	return NULL;
    }

    /* Initialize
       ---------- */
    sdim = subspace->Nor;
    qdim = subspace->Noc - sdim;
    result = MatAlloc(subspace->Field,vectors->Nor,qdim);

    /* Calculate the projection
       ------------------------ */
    FfSetNoc(subspace->Noc);
    tmp = FfAlloc(1);
    non_piv = subspace->PivotTable + subspace->Nor;
    for (i = 0; i < vectors->Nor; ++i)
    {
	int k;
	PTR q = MatGetPtr(result,i);
	FfCopyRow(tmp,MatGetPtr(vectors,i));
	FfCleanRow(tmp,subspace->Data,sdim,subspace->PivotTable);
	for (k = 0; k < qdim; ++k)
	    FfInsert(q,k,FfExtract(tmp,non_piv[k]));
    }
    SysFree(tmp);

    return result;
}



/*{{{ --------------------------------------------------------------------------
!function QAction "Action on a quotient"
!arg subspace
    Pointer to an invariant subspace.
!arg gen
    Matrix operating on the subspace.
!return
    Action of the generator on the quotient, or |NULL| on error.
!description
    Given a subspace $U\leq F^n$ and a matrix $a\in F^{n\times n}$ that maps 
    $U$ into $U$, this function calculates the action of the matrix on the 
    quotient $F^n/U$.
    As input, the function expects a basis of the subspace in |subspace|,
    which must be in echelon form, and the matrix operating on the
    subspace in |gen|. The result is a square matrix with $n-\dim(V)$ rows
    describing the action of $a$ on the quotient in a randomly chosen
    basis. 
    
    Before calculating the action, |QAction()| checks if the arguments
    are valid matrices, and if they are compatible. Both |subspace|
    and |gen| must be over the same field, have the same number of
    columns, and |gen| must be square.

    Note that |QAction()| is intended to be used with single matrices.
    If you want to compute the action of a whole set of matrices on the 
    same quotient, |Split()| may be more suitable.
!seealso SAction Split
-------------------------------------------------------------------------- }}}*/

Matrix_t *QAction(const Matrix_t *subspace, const Matrix_t *gen)

{
    int k;
    int dim, sdim, qdim;
    int *piv, *non_piv;
    PTR tmp;
    Matrix_t *action;

   /* Check arguments.
      ---------------- */
    if (!MatIsValid(subspace) || !MatIsValid(gen))
	return NULL;
    if (subspace->Noc != gen->Nor)
    {
	MTX_ERROR1("subspace and gen: %E",MTX_ERR_INCOMPAT);
	return NULL;
    }
    if (gen->Nor != gen->Noc)
    {
	MTX_ERROR1("gen: %E",MTX_ERR_NOTSQUARE);
	return NULL;
    }

    /* Initialize
       ---------- */
    dim = subspace->Noc;
    sdim = subspace->Nor;
    qdim = dim - sdim;
    if ((action = MatAlloc(subspace->Field,qdim,qdim)) == NULL)
	return NULL;

    /* Calculate the action on the quotient
       ------------------------------------ */
    FfSetNoc(dim);
    tmp = FfAlloc(1);
    piv = subspace->PivotTable;
    non_piv = piv + subspace->Nor;
    for (k = 0; k < qdim; ++k)
    {
	int l;
	PTR qx = MatGetPtr(action,k);
	FfCopyRow(tmp,MatGetPtr(gen,non_piv[k]));
	FfCleanRow(tmp,subspace->Data,sdim,piv);
	for (l = 0; l < qdim; ++l)
	    FfInsert(qx,l,FfExtract(tmp,non_piv[l]));
    }
    SysFree(tmp);

    return action;
}

