/* ============================= C MeatAxe ==================================
   File:        $Id: zcleanrow.c,v 1.2 2000/03/18 13:58:44 mringe Exp $
   Comment:     Reduce a matrix to semi echelon form.
   --------------------------------------------------------------------------
   (C) Copyright 1998 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */


#include "meataxe.h"
#include <stdlib.h>

MTX_DEFINE_FILE_INFO




/*{{{ --------------------------------------------------------------------------
!section kernel.ff.row
!function FfCleanRow "Clean row (Gaussian elimination)"
!index "Gaussian elimination"
!arg row
    The row to be cleaned.
!arg matrix
    Pointer to the matrix.
!arg nor
    Number of rows of the matrix.
!arg piv
    The pivot table.
!description
    This function performs a Gaussian elimination, i.e., it adds suitable
    multiples of the rows of |matrix| to |row| such that all pivot positions
    are zero. |piv| is the pivot table for |matrix|, where |piv[0]| is the
    pivot column of the first row of |matrix|. As usual, all indexes are
    0-based.  For example, if |matrix| is the unit matrix, then |piv[0]==0|,
    |piv[1]==1| and so on. The row size must have been set before calling
    this function.
!seealso @FfCleanRow
-------------------------------------------------------------------------- }}}*/

void FfCleanRow(PTR row, PTR matrix, int nor, const int *piv)

{
    int i;
    PTR x;

    for (i=0, x=matrix; i < nor; ++i, FfStepPtr(&x))
    {
        FEL f = FfExtract(row,piv[i]);
        if (f != FF_ZERO)
	    FfAddMulRow(row,x,FfNeg(FfDiv(f,FfExtract(x,piv[i]))));
    }
}


/*{{{ --------------------------------------------------------------------------
!function FfCleanRow2 "Clean row and record operations"
!arg row
    Pointer to row to be cleaned.
!arg mat
    Matrix to clean with.
!arg nor
    Number of rows.
!arg piv
    Pivot table for |matrix|.
!arg row2
    Pointer to row where the operations are recorded.
!return
    The return value is always $0$.
!description
    This function acts exactly as |FfCleanRow()|, but it stores a
    record of the operations performed in |row2|. |row2| must
    be a row of at least |nor| entries.  On return, |row2| contains the
    coefficients by which the rows of |mat| were multiplied and then
    subtracted from |row|.
    
    Note: Before calling |FfCleanRow2()|, the caller must initialize 
    |row2| to zero. Otherwise the results are undefined.
!seealso @FfCleanRow
-------------------------------------------------------------------------- }}}*/

void FfCleanRow2(PTR row, PTR mat, int nor, const int *piv, PTR row2)

{
    int i;
    PTR x;

    if (row2 == NULL || piv == NULL)
    {
	MTX_ERROR1("%E",MTX_ERR_BADARG);
	return;
    }
    for (i = 0, x = mat; i < nor; ++i, FfStepPtr(&x))
    {
	FEL f = FfExtract(row,piv[i]);
	if (f != FF_ZERO) 
	{
	    f = FfDiv(f,FfExtract(x,piv[i]));
	    FfAddMulRow(row,x,FfNeg(f));
	    FfInsert(row2,i,f);
	}
    }
}



/*{{{ --------------------------------------------------------------------------
!function FfCleanRowAndRepeat "Clean row and repeat operations"
!arg row
    Pointer to row to be cleaned.
!arg mat
    Matrix to clean with.
!arg nor
    Number of rows.
!arg piv
    Pivot table for |mat|.
!arg row2
    Pointer to the second row to be cleaned.
!arg mat2
    Matrix to the second matrix.
!return
    The return value is always $0$.
!description
    This function works like |FfCleanRow()|, but repeats all operations on
    a second row/matrix. 
!seealso @FfCleanRow 
-------------------------------------------------------------------------- }}}*/

void FfCleanRowAndRepeat(PTR row, PTR mat, int nor, const int *piv, 
    PTR row2, PTR mat2)

{
    int i;
    PTR x, x2;

#ifdef DEBUG
    if (row2 == NULL || piv == NULL || row2 == NULL || mat2 == NULL)
    {
	MTX_ERROR1("%E",MTX_ERR_BADARG);
	return;
    }
#endif
    for (i = 0, x = mat, x2 = mat2; i < nor; ++i, FfStepPtr(&x), FfStepPtr(&x2))
    {
	FEL f = FfExtract(row,piv[i]);
	if (f != FF_ZERO) 
	{
	    f = FfNeg(FfDiv(f,FfExtract(x,piv[i])));
	    FfAddMulRow(row,x,f);
	    FfAddMulRow(row2,x2,f);
	}
    }
}
