/* ============================= C MeatAxe ==================================
   File:        $Id: zqt.c,v 1.3 2001/04/04 17:59:38 mringe Exp $
   Comment:     Projection on quotient.
   --------------------------------------------------------------------------
   (C) Copyright 1999 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ========================================================================== */

#include "meataxe.h"
#include <string.h>
#include <stdlib.h>




/* ------------------------------------------------------------------
   Global data
   ------------------------------------------------------------------ */

MTX_DEFINE_FILE_INFO

static MtxApplicationInfo_t AppInfo = { 
"zqt", "Clean And Quotient", 
"SYNTAX\n"
"    zqt [-i] [<Subsp> <Matrix> <Quotient>]\n"
"\n"
"OPTIONS\n"
"    -i   Take only insignificant rows of <Matrix>. <Quotient> will be\n"
"         the action of <Matrix> on the quotient by <Subspace>.\n"
"\n"
"FILES\n"
"    <Subsp>    i  The invariant subspace, in semi-echelon form\n"
"    <Matrix>   i  The matrix, must have the same number of columns\n"
"    <Quotient> o  Insignificant columns of <Matrix>, cleaned with <Subspace>\n"
"\n"
};


static MtxApplication_t *App = NULL;
static int opt_i = 0;
static const char *mname, *sname, *oname;
static Matrix_t *Subspace = NULL;
static MtxFile_t *InputFile = NULL;
static MtxFile_t *OutputFile = NULL;
static PTR InputBuffer = NULL;
static PTR OutputBuffer = NULL;
static int QuotientDim;



/* ------------------------------------------------------------------
   init() - Process command line options and arguments
   ------------------------------------------------------------------ */


static int Init(int argc, const char **argv)

{
    App = AppAlloc(&AppInfo,argc,argv);
    if (App == NULL)
	return -1;
    opt_i = AppGetOption(App,"-i");
    if (AppGetArguments(App,3,3) < 0)
	return -1;
    sname = App->ArgV[0];
    mname = App->ArgV[1];
    oname = App->ArgV[2];
    return 0;
}




/* ------------------------------------------------------------------
   readfiles() - Read the input files, allocate tables
   ------------------------------------------------------------------ */

static int ReadFiles()

{
    /* Read the subspace, and build the pivot table.
       --------------------------------------------- */
    Subspace = MatLoad(sname);
    if (Subspace == NULL)
	return 1;
    if (MatPivotize(Subspace) < 0)
    {
	MTX_ERROR2("%s: %E",mname,MTX_ERR_NOTECH);
	return 1;
    }

    /* Open the input file, check compatibility, and allocate buffer.
       -------------------------------------------------------------- */
    InputFile = MfOpen(mname);
    if (InputFile == NULL)
	return 1;
    if (InputFile->Field != Subspace->Field || InputFile->Noc != Subspace->Noc)
    {
	MTX_ERROR3("%s and %S: %E",sname,mname,MTX_ERR_INCOMPAT);
	return 1;
    }
    if (opt_i && InputFile->Nor != InputFile->Noc) 
    {
	MTX_ERROR2("%s: %E",mname,MTX_ERR_NOTSQUARE);
	return 1;
    }
    FfSetNoc(InputFile->Noc);
    InputBuffer = FfAlloc(1);
    QuotientDim = Subspace->Noc - Subspace->Nor;

    /* Open the output file and allocate buffer.
       ----------------------------------------- */
    OutputFile = MfCreate(oname,FfOrder,opt_i ? QuotientDim : InputFile->Nor,
	QuotientDim);
    OutputBuffer = FfAlloc(1);

    return 0;
}





/* ------------------------------------------------------------------
   IsPivot() - Find out if a given column is significant or not.
   ------------------------------------------------------------------ */

static int IsPivot(int i)

{
    int k;
    int *piv = Subspace->PivotTable;
    int sdim = Subspace->Nor;
    for (k = 0; k < sdim; ++k)
    {
	if (i == piv[k])
	    return 1;
    }
    return 0;
}




/* ------------------------------------------------------------------
   doit()
   ------------------------------------------------------------------ */

static int doit()

{
    int i;
    int *non_pivot = Subspace->PivotTable + Subspace->Nor;

    for (i = 0; i < InputFile->Nor; ++i)
    {
	int k;

	/* Read one row from the input file.
	   --------------------------------- */
	if (MfReadRows(InputFile,InputBuffer,1) != 1)
	{
	    MTX_ERROR("Error reading vector");
	    return 1;
	}

	/* When calculating the action, take only `insignifican' rows.
	   ----------------------------------------------------------- */
	if (opt_i && IsPivot(i))
	    continue;

	/* Clean and extract insignificant columns.
	   ---------------------------------------- */
	FfCleanRow(InputBuffer,Subspace->Data,Subspace->Nor,
	    Subspace->PivotTable);
	FfSetNoc(QuotientDim);
	FfMulRow(OutputBuffer,FF_ZERO);
	for (k = 0; k < QuotientDim; ++k)
	    FfInsert(OutputBuffer,k,FfExtract(InputBuffer,non_pivot[k]));

	/* Write the output row.
	   --------------------- */
	if (MfWriteRows(OutputFile,OutputBuffer,1) != 1)
	{
	    MTX_ERROR("Error writing vector");
	    return 1;
	}

    }

    return 0;
}



static void Cleanup()

{
    if (Subspace != NULL)
	MatFree(Subspace);
    if (InputFile != NULL)
	MfClose(InputFile);
    if (OutputFile != NULL)
	MfClose(OutputFile);
    AppFree(App);
}


/* ------------------------------------------------------------------
   main()
   ------------------------------------------------------------------ */

int main(int argc, const char **argv)

{
    int rc;
    if (Init(argc,argv) != 0)
    {
	MTX_ERROR("Initialization failed");
	return 1;
    }
    if (ReadFiles() != 0)
    {
	MTX_ERROR("Error reading files");
	return 1;
    }
    rc = doit();
    Cleanup();
    return rc;
}



/*{{{ --------------------------------------------------------------------------
!section apps.general
!program zqt "Clean and quotient"
!synopsis 
    zqt [-i] <Subsp> <Matrix> <Quot>
!option -i
    Calculate the action on the quotient.
!arg <Subsp>
    Input file: the subspace.
!arg <Matrix>
    Input file: the matrix.
!arg <Quot>
    Output file: quotient.
!description
This program reads in a subspace from <Subsp> and applies the
canonical map to its quotient on the matrix <Matrix>, the result
being written out to <Quot>.
<Subsp> should be a matrix in semi-echelon form, and the two
input matrices must have the same field parameter and the same number
of columns. If this is not the case the program stops with an error
message.

Otherwise the program reads in <Subsp>, builds a table of pivot
columns and then proceeds, row by row, through <Matrix>. For
each row, the significant entries are zeroized by adding the correct
multiple of rows of <Subsp>. The insignificant columns are then
extracted and written out to <Quot>. Hence
- <Subsp> has A rows, B columns and is in echelon form,
- <Matrix> has C rows, B columns and is otherwise
  arbitrary, and
- <Quot> has C rows and B-A columns.
In other words, the program calculates the projection of <Matrix>
onto the B-A dimensional quotient space defined by <Subsp>.  
If the `-i' option is used, ZQT calculates the action of
<Matrix> on the quotient by <Subsp>. This is done
by projecting the matrix as explained above, and taking only the
`insignificant' rows. Insignificant rows are defined by treating
the pivot table as a table of rows rather than columns.

Example: Let `spc' be an invariant subspace and `z1' an
algebra element (a square matrix). Then, after
||    zqt -i spc z1 q1
`q1' contains the action of `z1' on the quotient by `spc'.

Another, less obvious use of ZQT is to condense a matrix
representation. First, find an element E of the group algebra with
stable rank, i.e.\ rank(E*E) = rank(E). This can be found by taking
any element F of the group algebra and raising it to higher powers
until the rank stabilizes. We may then condense onto the kernel of E
as follows
||    zef E X        X is the echelon form of Image(E)
||    znu E Y        Y is the kernel of E
||    zqt X Y Z      calculate the canonical projection of Y ...
||    ziv Z T        ... and adjust Y so that the canonical ...
||    zmu T Y Y1     ... projection of Y1 is the identity
||    zmu Y1 Z1 T1   calculate KZ1 = condensed Z1
||    zqt X T1 KZ1
||    zmu Y1 Z2 T1   calculate KZ2 = condensed Z2
||    zqt X T1 KZ2

*Limitations
It is not completely checked that <Subsp> is in echelon form.

The first matrix (<Subsp>) and one row of both <Matrix> and
<Subsp> must fit into memory.
-------------------------------------------------------------------------- }}}*/


